/**************************************************************************
***    
*** Copyright (c) 2004 Regents of the University of Michigan,
***               Hayward H. Chan and Igor L. Markov
***
***  Contact author(s): hhchan@umich.edu, imarkov@umich.edu
***  Original Affiliation:   EECS Department, 
***                          The University of Michigan,
***                          Ann Arbor, MI 48109-2122
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/
#include "debug.h"
#include "datastrbacksoftst.h"
#include "datastrfrontsoftst.h"
#include "datastrhiersoftst.h"
#include "enginesoftst.h"
#include "enginehiersoftst.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <algorithm>
using namespace std;

// --------------------------------------------------------
int Debug::count = 0;
string Debug::filePrefix("");
string Debug::fileSuffix("");
const string Debug::TEMP_FILE_NAME("out");
// --------------------------------------------------------
void Debug::initialize(const string& nName)
{
   stringstream old(nName);
   char ch;

   filePrefix = "";
   old.get(ch);
   while (ch != '.' && old.good())
   {
      filePrefix += ch;
      old.get(ch);
   }

   fileSuffix = "";
   fileSuffix += '.';
   old.get(ch);
   if (old.good())
      while (old.good())
      {
         fileSuffix += ch;
         old.get(ch);
      }
   else
      fileSuffix = ".ps";
}
// --------------------------------------------------------    
void Debug::OutputBoundary(int format,
                           ostream& outs,
                           const BoundaryType& boundary)
{
   if (boundary.size() < 3)
   {
      cout << "ERROR: in OutputBoundary(), boundary has less than "
           << "3 elements. " << endl;
      exit(1);
   }

   int ptr = boundary.size()-1;
   while (boundary[ptr].xCoord() >= Point::INFTY)
      ptr--;
   double xEnd = boundary[ptr].xCoord() * 1.1;

   ptr = 0;
   while (boundary[ptr].yCoord() >= Point::INFTY)
      ptr++;
   double yEnd = boundary[ptr].yCoord() * 1.1;

   stringstream ss;
   switch (format)
   {
   case GNUPLOT: // gnuplot format
      ss << filePrefix << "-" << count << fileSuffix;
      count++;
      
      cout << "Writing gnuplot script to create file: "
           << ss.str() << endl;           
      GnuplotHeading(outs, ss.str());

      outs << "set xtics" << endl;
      outs << "set ytics" << endl;
      outs << "set title \"boundary " << (count-1) << "\"" << endl;
      outs << "plot [0:" << xEnd << "][0:" << yEnd << "] ";
      outs << "\"-\" notitle with linesp lt 1 pt 3" << endl;
      for (int i = 0; i < boundary.size(); i++)
         outs << min(boundary[i].xCoord(), xEnd) << " "
              << min(boundary[i].yCoord(), yEnd) << endl;
      outs << "END" << endl;
      break;
   case CONSOLE:      
      outs << "area: " << boundary.min_area() << endl;
      for (int i = 0; i < boundary.size(); i++)
         outs << boundary[i].xCoord() << " "
              << boundary[i].yCoord() << endl;
      break;
   }
}
// --------------------------------------------------------
void Debug::GnuplotHeading(ostream& outs,
                           const string& filename)
{
   outs << "set noxtics" << endl;
   outs << "set noytics" << endl;
   outs << "set nox2tics" << endl;
   outs << "set noy2tics" << endl;
   outs << "set size ratio -1" << endl;
   outs << "set terminal postscript landscape solid" << endl;
   outs << "set output \"";
   outs << filename;
   outs << "\"" << endl;
}
// --------------------------------------------------------
void Debug::ReadCurve(istream& ins,
                      vector<Point>& curve)
{
   int numPts;
   ins >> numPts;

   curve.clear();
   for (int j = 0; j < numPts; j++)
   {
      double x, y;
      if (!ins.good())
      {
         cout << "ERROR: error before reading point["
              << j << "]" << endl;
         exit(1);
      }
      ins >> x >> y;

      if (j == 0)
         curve.push_back(Point(x, Point::INFTY));
      curve.push_back(Point(x,y));
      if (j == numPts-1)
         curve.push_back(Point(Point::INFTY, y));
   }
}      
// --------------------------------------------------------
void Debug::PlotCurves(ostream& outs,
                       const vector< vector<Point> >& curves)
{
   stringstream ss;
   ss << filePrefix << "-" << count << fileSuffix;
   count++;

   GnuplotHeading(outs, ss.str());
      
   cout << "Writing gnuplot script to create file: "
        << ss.str() << endl;           

   outs << "set xtics" << endl;
   outs << "set ytics" << endl;
   outs << "set title \"" << ss.str() << "\"" << endl;

   double xEnd = 0;
   double yEnd = 0;

   for (unsigned int i = 0; i < curves.size(); i++)
      for (unsigned int j = 1; j < curves[i].size()-1; j++)
      {
         xEnd = max(xEnd, curves[i][j].xCoord());
         yEnd = max(yEnd, curves[i][j].yCoord());
      }

   xEnd *= 1.1;
   yEnd *= 1.1;
   
   outs << "plot [0:" << xEnd << "][0:" << yEnd << "] ";
   for (unsigned int i = 0; i < curves.size(); i++)
   {
      outs << "\"-\" with linesp lw " << i << " pt 3";
      if (i < curves.size()-1)
         outs << ", ";
   }
   outs << endl;
   
   for (unsigned int i = 0; i < curves.size(); i++)
   {
      for (unsigned int j = 0; j < curves[i].size(); j++)
         outs << min(curves[i][j].xCoord(), xEnd) << " "
              << min(curves[i][j].yCoord(), yEnd) << endl;
      outs << "END" << endl;
   }
}
// --------------------------------------------------------
void Debug::OutputNode(ostream& outs,
                       const SoftNode& node)
{
   outs << "sign: ";
   if (node.sign == BoundaryType::PLUS)
      outs << "+" << endl;
   else if (node.sign == BoundaryType::STAR)
      outs << "*" << endl;
   else
      outs << node.sign << endl;

   outs << "BLBlock: " << node.BLBlock << endl;
   outs << "TRblblock: " << node.TRblblock << endl;
   outs << "blkArea: " << node.blkArea << endl;
   outs << "minArea: " << node.minArea << endl;
   outs << "minDspc: " << node.minDeadspace << endl;
}
// --------------------------------------------------------
void Debug::OutputSoftSTree(ostream& outs,
                            const SoftSTree& st)
{
   cout.setf(ios::fixed);
   cout.precision(2);
   
   cout << "balance: " << st.balance() << endl;
   cout << "perSize: " << st.perSize() << endl;
   cout << "d_Space: " << st.deadspace() << endl;
   cout << "b__Area: " << st.total_area() - st.deadspace() << endl;
   cout << "t__Area: " << st.total_area() << endl;

   cout << "expression: ";
   for (int i = 0; i < st.expression_size(); i++)
      if (st.expression(i) == SoftSTree::PLUS)
         cout << "+ ";
      else if (st.expression(i) == SoftSTree::STAR)
         cout << "* ";
      else 
         cout << st.expression(i) << " ";
   cout << endl;

   cout.precision(0);
   cout << "-----buffer-----" << endl;
   cout << " sign : ";
   for (int i = 0; i < st.buffer_size(); i++)
      if (st.buffer(i).sign == SoftSTree::PLUS)
         cout << setw(5) << '+';
      else if (st.buffer(i).sign == SoftSTree::STAR)
         cout << setw(5) << '*';
      else 
         cout << setw(5) << st.buffer(i).sign;
   cout << endl;
   cout << " blblk: ";
   for (int i = 0; i < st.buffer_size(); i++)
      cout << setw(5) << st.buffer(i).BLBlock;
   cout << endl;
   cout << " trblb: ";
   for (int i = 0; i < st.buffer_size(); i++)
      cout << setw(5) << st.buffer(i).TRblblock;
   cout << endl;
   cout << " blk_a: ";
   for (int i = 0; i < st.buffer_size(); i++)
      cout << setw(5) << st.buffer(i).blkArea;
   cout << endl;
   cout << " min_a: ";
   for (int i = 0; i < st.buffer_size(); i++)
      cout << setw(5) << st.buffer(i).minArea;
   cout << endl;
   cout << " min_d: ";
   for (int i = 0; i < st.buffer_size(); i++)
      cout << setw(5) << st.buffer(i).minDeadspace;
   cout << endl;

   cout << endl;
   cout << "-----storage-----" << endl;
   cout << " sign : ";
   for (int i = 0; i < st.storage_size(); i++)
      if (st.storage(i).sign == SoftSTree::PLUS)
         cout << setw(5) << '+';
      else if (st.storage(i).sign == SoftSTree::STAR)
         cout << setw(5) << '*';
      else 
         cout << setw(5) << st.storage(i).sign;
   cout << endl;
   cout << " blblk: ";
   for (int i = 0; i < st.storage_size(); i++)
      cout << setw(5) << st.storage(i).BLBlock;
   cout << endl;
   cout << " trblb: ";
   for (int i = 0; i < st.storage_size(); i++)
      cout << setw(5) << st.storage(i).TRblblock;
   cout << endl;
   cout << " blk_a: ";
   for (int i = 0; i < st.storage_size(); i++)
      cout << setw(5) << st.storage(i).blkArea;
   cout << endl;
   cout << " min_a: ";
   for (int i = 0; i < st.storage_size(); i++)
      cout << setw(5) << st.storage(i).minArea;
   cout << endl;
   cout << " min_d: ";
   for (int i = 0; i < st.storage_size(); i++)
      cout << setw(5) << st.storage(i).minDeadspace;
   cout << endl;

   cout << endl;
   cout << "-----BLOCK_NODES-----" << endl;
   cout << " sign : ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      if (st.BLOCK_NODES[i].sign == SoftSTree::PLUS)
         cout << setw(5) << '+';
      else if (st.BLOCK_NODES[i].sign == SoftSTree::STAR)
         cout << setw(5) << '*';
      else 
         cout << setw(5) << st.BLOCK_NODES[i].sign;
   cout << endl;
   cout << " blblk: ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      cout << setw(5) << st.BLOCK_NODES[i].BLBlock;
   cout << endl;
   cout << " trblb: ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      cout << setw(5) << st.BLOCK_NODES[i].TRblblock;
   cout << endl;
   cout << " blk_a: ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      cout << setw(5) << st.BLOCK_NODES[i].blkArea;
   cout << endl;
   cout << " min_a: ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      cout << setw(5) << st.BLOCK_NODES[i].minArea;
   cout << endl;
   cout << " min_d: ";
   for (unsigned int i = 0; i < st.BLOCK_NODES.size(); i++)
      cout << setw(5) << st.BLOCK_NODES[i].minDeadspace;
   cout << endl;
}
// --------------------------------------------------------
void Debug::OutputSoftSliceRecord(ostream& outs,
                                  const SoftSliceRecord& ssr)
{
   cout << "sign: ";
   for (unsigned int i = 0; i < ssr.expression.size(); i++)
      if (ssr.expression[i] == SoftSTree::PLUS)
         cout << setw(5) << '+';
      else if (ssr.expression[i] == SoftSTree::STAR)
         cout << setw(5) << '*';
      else if (ssr.expression[i] == SoftSTree::BOTH)
         cout << setw(5) << '-';
      else
         cout << setw(5) << ssr.expression[i];
   cout << endl;
   cout << "min_a: " << ssr.minArea << endl;
   cout << "min_d: " << ssr.minDeadspace << endl;
}
// --------------------------------------------------------
void Debug::CheckSTree(const SoftSTree& sst)
{
   for (int i = 0; i < sst.buffer_size(); i++)
      if (sst.buffer(i).boundary.size() < 5)
      {
         cout << "boundary error starts here." << endl;
         for (int j = 0; j < sst.expression_size(); j++)
            if (sst.expression(j) == SoftSTree::PLUS)
               cout << "+ ";
            else if (sst.expression(j) == SoftSTree::STAR)
               cout << "* ";
            else
               cout << sst.expression(j) << " ";         
         cout << endl;
         
         cout << "pre-buffer: " << endl;
         for (int j = 0; j < sst.buffer_size(); j++)
         {
            Debug::OutputNode(cout, sst.buffer(j));
            cout << "size(): " << sst.buffer(j).boundary.size() << endl;
            
            cout << "x: ";
            for (int k = 0; k < sst.buffer(j).boundary.size(); k++)
               if (sst.buffer(j).boundary[k].xCoord() >= Point::INFTY)
                  cout << setw(7) << "---";
               else
                  cout << setw(7) << sst.buffer(j).boundary[k].xCoord();
            cout << endl;
            cout << "y: ";
            for (int k = 0; k < sst.buffer(j).boundary.size(); k++)
               if (sst.buffer(j).boundary[k].yCoord() >= Point::INFTY)
                  cout << setw(7) << "---";
               else
                  cout << setw(7) << sst.buffer(j).boundary[k].yCoord();
            cout << endl << endl;
            
         }
         cout << endl;
         cout << "pre-storage: " << endl;
         for (int j = 0; j < sst.storage_size(); j++)
         {
            Debug::OutputNode(cout, sst.storage(j));
            cout << "size(): " << sst.storage(j).boundary.size() << endl;

            cout << "x: ";
            for (int k = 0; k < sst.storage(j).boundary.size(); k++)
               if (sst.storage(j).boundary[k].xCoord() >= Point::INFTY)
                  cout << setw(7) << "---";
               else
                  cout << setw(7) << sst.storage(j).boundary[k].xCoord();
            cout << endl;
            cout << "y: ";
            for (int k = 0; k < sst.storage(j).boundary.size(); k++)
               if (sst.storage(j).boundary[k].yCoord() >= Point::INFTY)
                  cout << setw(7) << "---";
               else
                  cout << setw(7) << sst.storage(j).boundary[k].yCoord();
            cout << endl << endl;
         }         
         exit(1);
      }
}  
// --------------------------------------------------------

// -----debugging program-----
void DebugSTree(int argc, char *argv[]);
void DebugEval(int argc, char *argv[]);
void DebugBrute(int argc, char *argv[]);
void DebugReadTxt(int argc, char *argv[]);
void DebugPlus(int argc, char *argv[]);
void DebugHier(int argc, char *argv[]);

int main(int argc, char *argv[])
{
   vector< vector<Point> > curves;   

   if (!strcmp(argv[2], "-i"))
   {
      int numPts, numCurves;

      cout << "# curves ->";
      cin >> numCurves;
   
      cout << "# points (excluding end points->";
      cin >> numPts;

      curves.resize(numCurves);
      for (int i = 0; i < numCurves; i++)
         for (int j = 0; j < numPts; j++)
         {
            double xCoord, yCoord;
            cout << "point[" << i << "][" << j << "] ->";
            cin >> xCoord >> yCoord;
         
            if (j == 0)
               curves[i].push_back(Point(xCoord, Point::INFTY));
         
            curves[i].push_back(Point(xCoord, yCoord));
            if (j == numPts-1)
               curves[i].push_back(Point(Point::INFTY, yCoord));
         }
   }
   else if (!strcmp(argv[2], "-f") ||
            !strcmp(argv[2], "--interpolate"))
   {
      int numCurves;     
      cout << "# curves ->";
      cin >> numCurves;

      curves.resize(numCurves);
      for (int i = 0; i < numCurves; i++)
      {
         string filename;
         cout << "filename[" << i << "]->";
         cin >> filename;
         
         ifstream infile;
         infile.open(filename.c_str());

         int numPts;
         infile >> numPts;

         for (int j = 0; j < numPts; j++)
         {
            double x, y;
            if (!infile.good())
            {
               cout << "ERROR: error before reading point " << j
                    << " of curve " << i << endl;
               exit(0);
            }
            infile >> x >> y;

            if (j == 0)
               curves[i].push_back(Point(x, Point::INFTY));
            curves[i].push_back(Point(x, y));
            if (j == numPts-1)
               curves[i].push_back(Point(Point::INFTY, y));
         }
         infile.close();
      }
   }
   if (!strcmp(argv[2], "--plus"))
      DebugPlus(argc, argv);
   
  

   if (!strcmp(argv[2], "-i") ||
       !strcmp(argv[2], "-f"))
   {
      ofstream outfile;
      outfile.open(Debug::TEMP_FILE_NAME.c_str());
      Debug::initialize(argv[1]);
      Debug::PlotCurves(outfile, curves);

      cout << "Point::INFTY: " << Point::INFTY << endl;
      cout << "min_area[" << 0 << "]: " << BoundaryType(curves[0]).min_area() << endl;
      cout << "min_area[" << 1 << "]: " << BoundaryType(curves[1]).min_area() << endl;
      cout << "min_area[" << 2 << "]: " << BoundaryType(curves[2]).min_area() << endl;
      outfile.close();
   }

   if (!strcmp(argv[2], "--blockinfo"))
   {
      ifstream infile;

      infile.open(argv[3]);
      if (!infile.good())
      {
         cout << "ERROR: cannot open file: " << argv[3] << endl;
         exit(1);
      }
      BlockInfoType blockinfo(BlockInfoType::SOFT, infile);

      Debug::initialize(argv[1]);
      int blocknum = blockinfo.BLOCK_NUM();
      for (int i = 0; i < blocknum; i++)
      {
         ofstream outfile;
         stringstream ss;

         ss << "out" << i;
         outfile.open(ss.str().c_str());
         Debug::OutputBoundary(Debug::GNUPLOT, outfile, blockinfo[i]);
         cout << "min_area_x: " << blockinfo[i].min_point().xCoord()
              << " min_area_y: " << blockinfo[i].min_point().yCoord()
              << " min_area: " << blockinfo[i].min_point().area() << endl;
         outfile.close();
      }
      cout << "total_area: " << blockinfo.total_area() << endl;
   }

   if (!strcmp(argv[2], "--node"))
   {
      ifstream infile;

      infile.open(argv[3]);
      if (!infile.good())
      {
         cout << "ERROR: cannot open file: " << argv[3] << endl;
         exit(1);      
      }
      BlockInfoType blockinfo(BlockInfoType::SOFT, 40, infile);

      char ch = '\0';
      vector<SoftNode> nodes;

      for (int i = 0; i < blockinfo.BLOCK_NUM(); i++)
         nodes.push_back(SoftNode(i, blockinfo[i]));

      Debug::initialize(argv[1]);
      while (ch != 'q' && cin.good())
      {
         cout << "action (s)tar and (p)lus->" << endl;
         cin >> ch;

         int a, b;
         switch (ch)
         {
         case 's':
            cout << "nodes 0-" << (nodes.size()-1) << "->";
            cin >> a >> b;
            nodes.push_back(SoftNode(BoundaryType::STAR, nodes[a], nodes[b]));
            cout << nodes[a] << endl;
            cout << nodes[b] << endl;
            cout << nodes.back() << endl;
            break;
         case 'p':
            cout << "nodes 0-" << (nodes.size()-1) << "->";
            cin >> a >> b;
            nodes.push_back(SoftNode(BoundaryType::PLUS, nodes[a], nodes[b]));
            cout << nodes[a] << endl;
            cout << nodes[b] << endl;
            cout << nodes.back() << endl;
            break;
         }

         ofstream outfile;
         static int count = 0;
         stringstream ss;
         ss << "out" << count;
         count++;
         
         outfile.open(ss.str().c_str());
         if (!outfile.good())
         {
            cout << "cannot open: " << ss.str() << endl;
            exit(1);
         }
         Debug::OutputBoundary(Debug::GNUPLOT, outfile, nodes.back().boundary);
         outfile.close();

         for (int i = 1; i < nodes.back().boundary.size()-1; i++)
            if (nodes.back().boundary[i].xCoord() >= Point::INFTY ||
                nodes.back().boundary[i].yCoord() >= Point::INFTY)
            {
               cout << "Duplicated infty[" << i << "]: x: "
                    << nodes.back().boundary[i].xCoord()
                    << " y: " << nodes.back().boundary[i].yCoord() << endl;               
            }
      }
   }

   if (!strcmp(argv[2], "--interpolate"))
   {
      BoundaryType bt(curves[0]);

//      Debug::OutputBoundary(Debug::CONSOLE, cout, bt);
      cout << "size:  " << bt.size() << endl;
      cout << "minX: " << bt[0].xCoord() << " minY: "
           << bt[bt.size()-1].yCoord() << endl;
      while (cin.good())
      {
         double coord;
         char ch;

         cout << "provide x or y ->";
         cin >> ch;

         switch (ch)
         {
         case 'x':
            cout << "xCoord ->";
            cin >> coord;
            cout << "y: " << bt.getY(coord) << endl;
            break;
         case 'y':
            cout << "yCoord ->";
            cin >> coord;
            cout << "x: " << bt.getX(coord) << endl;
            break;
         }
      }
   }

   if (!strcmp(argv[2], "--stree"))
      DebugSTree(argc, argv);

   if (!strcmp(argv[2], "--eval"))
      DebugEval(argc, argv);

   if (!strcmp(argv[2], "--brute") ||
       !strcmp(argv[2], "--bound"))
      DebugBrute(argc, argv);

   if (!strcmp(argv[2], "--readtxt"))
      DebugReadTxt(argc, argv);

   if (!strcmp(argv[2], "--hier"))
      DebugHier(argc, argv);
   return 0;
}
// --------------------------------------------------------
void DebugSTree(int argc, char *argv[])
{
   ifstream infile;

   infile.open(argv[1]);
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << argv[1] << endl;
      exit(1);
   }

   int format = -1;
   if (!strcmp(argv[3], "--soft"))
      format = BlockInfoType::SOFT;
   else if (!strcmp(argv[3], "--txt"))
      format = BlockInfoType::TXT;
   
   BlockInfoType blockinfo(format, infile);
   SoftSTree st(blockinfo);
   SoftSliceRecord *ptr = NULL;

   Debug::OutputSoftSTree(cout, st);
   while (cin.good())
   {
      char ch;
      int sign;
      char sign_ch;
      
      cout << "action: " << endl;
      cout << " push operan(d), push operato(r), p(o)p operand, (p)op operator, (s)ave record->";
      cin >> ch;

      switch (ch)
      {
      case 'd':
         cout << "block ->";
         cin >> sign;
         st.push_operand(sign);
         break;

      case 'r':
         cout << "operator ->";
         cin >> sign_ch;

         switch (sign_ch)
         {
         case '+':
            sign = SoftSTree::PLUS;
            break;
         case '*':
            sign = SoftSTree::STAR;
            break;
         }
         
         if (st.can_push_operator(sign))
         st.push_operator(sign);
         else
            cout << "Cannot push the chosn operator." << endl;
         break;
         
      case 'o':
         if ((st.expression_top() == SoftSTree::PLUS) ||
             (st.expression_top() == SoftSTree::STAR))
            cout << "the top of expression is operator." << endl;
         else
            st.pop_operand();
         break;
         
      case 'p':
         if ((st.expression_top() == SoftSTree::PLUS) ||
             (st.expression_top() == SoftSTree::STAR))
            st.pop_operator();
         else
            cout << "the top of expression is operand." << endl;
         break;
         
      case 'q':
         exit(0);
         break;
        
      default:
         cout << "invalid command: " << ch << endl;
         break;
      }
      
      if (ch == 's')
      {
         ptr = new SoftSliceRecord(st);
         cout << "-----SoftSliceRecord-----" << endl;
         Debug::OutputSoftSliceRecord(cout, *ptr);
         cout << "ssr.size(): " << ptr->boundary.size() << endl;

         ofstream outfile2;
         outfile2.open("DEBUG/bdy");
         Debug::OutputBoundary(Debug::GNUPLOT, cout, ptr->boundary);
         Debug::OutputBoundary(Debug::GNUPLOT, outfile2, ptr->boundary);         
         
         SoftPacking spk(*ptr, blockinfo);   
         cout << "after spk" << endl;
         spk.output(cout);

         ofstream outfile;
         outfile.open("DEBUG/temp.bbb");
         spk.output(outfile);
         outfile.close();
         delete ptr;
      }
         
      Debug::OutputSoftSTree(cout, st);
      for (int i = 0; i < st.buffer_size(); i++)
         cout << "[" << i << "] numPts: " << st.buffer(i).boundary.size() << endl;
      cout << endl;
   }
}
// --------------------------------------------------------
void DebugEval(int argc, char *argv[])
{
   ifstream infile;
   ofstream outfile;

   infile.open(argv[1]);
   if (!infile.good())
   {
      cout << "Cannot open file: " << argv[1] << endl;
      exit(1);
   }

   outfile.open(argv[3]);
   if (!infile.good())
   {
      cout << "Cannot open file: " << argv[3] << endl;
      exit(1);
   }

   int format = -1;
   if (!strcmp(argv[4], "--txt"))
      format = BlockInfoType::TXT;
   else if (!strcmp(argv[4], "--soft"))
      format = BlockInfoType::SOFT;
   
   BlockInfoType blockinfo(format, infile);
   SoftSTree st(blockinfo);

   int expr_size = 2 * blockinfo.BLOCK_NUM() - 1;
   vector<int> expression(expr_size);

   cout << "expr: ";
   for (int i = 0; i < expr_size; i++)
   {
      int sign;
      infile >> sign;

      if (sign == blockinfo.BLOCK_NUM() ||
          sign == SoftSTree::PLUS)
      {
         expression[i] = SoftSTree::PLUS;
         cout << "+ ";
      }
      else if (sign == blockinfo.BLOCK_NUM()+1 ||
               sign == SoftSTree::STAR)
      {
         expression[i] = SoftSTree::STAR;
         cout << "* ";
      }
      else if (sign < blockinfo.BLOCK_NUM())
      {
         expression[i] = sign;
         cout << sign << " ";
      }
      else
      {
         cout << "Invalid sign." << endl;
         exit(1);
      }      
   }
   cout << endl;

   for (int i = 0; i < expr_size; i++)
   {
      int sign = expression[i];
      switch (sign)
      {
      case SoftSTree::PLUS:
      case SoftSTree::STAR:
      case SoftSTree::BOTH:
         if (!strcmp(argv[5], "--explicit"))
            st.push_operator(sign);
         else if (!strcmp(argv[5], "--implicit"))
            st.push_operator(SoftSTree::BOTH);
         else
         {
            cout << "invalid mode: " << argv[5] << endl;
            exit(1);
         }
         break;
      default:
         st.push_operand(sign);
         break;
      }
   }

   Debug::initialize("bound.ps");
   SoftSliceRecord ssr(st);
   SoftPacking spk(ssr, blockinfo);
   spk.output(outfile);
}
// --------------------------------------------------------
void DebugBrute(int argc, char *argv[])
{
   ifstream infile;
   ofstream outfile;

   infile.open(argv[1]);
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << argv[1] << endl;
      exit(1);
   }

   outfile.open(argv[3]);
   if (!outfile.good())
   {
      cout << "ERROR: cannot open file: " << argv[3] << endl;
      exit(1);
   }

   int format = -1;
   if (!strcmp(argv[4], "--txt"))
      format = BlockInfoType::TXT;
   else if (!strcmp(argv[4], "--soft"))
      format = BlockInfoType::SOFT;
   
   BlockInfoType blockinfo(format, infile);
   SoftEngineType engine(blockinfo);

   SoftSliceRecordList *ssr_ptr = NULL;

   int mode = -1;
   if (!strcmp(argv[5], "--explicit"))
      mode = SoftEngineType::EXPLICIT;
   else if (!strcmp(argv[5], "--implicit"))
      mode = SoftEngineType::IMPLICIT;

   int algo = -1;
   if (!strcmp(argv[2], "--brute"))
      algo = SoftEngineType::BRUTE;
   else if (!strcmp(argv[2], "--bound"))
      algo = SoftEngineType::B_BOUND;   
       
   if (!strcmp(argv[2], "--brute"))
      ssr_ptr = engine(algo, mode);
   else if (!strcmp(argv[2], "--bound"))
      ssr_ptr = engine(algo, mode);

   cout << "last records: " << endl;
   SoftPacking(ssr_ptr->last(), blockinfo).output(outfile);
   SoftPacking(ssr_ptr->last(), blockinfo).output(cout);

   outfile.close();
   
//    outfile.open(argv[5]);
//    if (!outfile.good())
//    {
//       cout << "ERROR: cannot open file: " << argv[5] << endl;
//       exit(1);
//    }
   
//    cout << "size: " << ssr_ptr->last().boundary.size() << endl;
//    for (int i = 0; i < ssr_ptr->last().boundary.size(); i++)
//       cout << "[" << i << "] x: " << ssr_ptr->last().boundary[i].xCoord()
//            << "y: " << ssr_ptr->last().boundary[i].yCoord() << endl;
//    cout << endl;
//    Debug::OutputBoundary(Debug::GNUPLOT, outfile, ssr_ptr->last().boundary);
   
   delete ssr_ptr;
}
// --------------------------------------------------------
void DebugReadTxt(int argc, char *argv[])
{
   ifstream infile;

   infile.open(argv[1]);
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << argv[1] << endl;
      exit(1);
   }

   BlockInfoType blockinfo(BlockInfoType::TXT, infile);

   cout.setf(ios::fixed);
   cout.precision(2);
   for (int i = 0; i < blockinfo.size(); i++)
      cout << "[" << i << "] x: " << blockinfo[i][1].xCoord() << " "
           << "y: " << blockinfo[i][1].yCoord() << " "
           << "a: " << blockinfo[i].min_area() << endl;
   cout << endl;
   cout << "total_area: " << blockinfo.block_area() << endl;
   exit(0);
}
// --------------------------------------------------------
void DebugPlus(int argc, char *argv[])
{
   vector< vector<Point> > curves;
   curves.resize(4);
   string filename(argv[3]);

   ifstream infile;
   infile.open(filename.c_str());
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << filename << endl;
      exit(1);
   }
   Debug::ReadCurve(infile, curves[0]);
   infile.close();

   filename = argv[4];
   infile.open(filename.c_str());
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << filename << endl;
      exit(1);
   }
   Debug::ReadCurve(infile, curves[1]);
   infile.close();

   BoundaryType b0a(BoundaryType::PLUS, BoundaryType(curves[0]), BoundaryType(curves[1]));
   BoundaryType b0b(BoundaryType::STAR, BoundaryType(curves[0]), BoundaryType(curves[1]));
   BoundaryType b1(BoundaryType::BOTH, BoundaryType(curves[0]), BoundaryType(curves[1]));
   BoundaryType b2(BoundaryType::BOTH, BoundaryType(curves[1]), BoundaryType(curves[0]));

   vector<Point> block_A(curves[0]);
   vector<Point> block_B(curves[1]);
   curves[0] = b0a.curve();
   curves[1] = b0b.curve();
   curves[2] = b1.curve();
   curves[3] = b2.curve();
   
   for (unsigned int j = 0; j < min(curves[2].size(), curves[3].size()); j++)
      if (curves[2][j].xCoord() != curves[3][j].xCoord() ||
          curves[2][j].yCoord() != curves[3][j].yCoord())
      {
         cout << "ERROR: the two curves disagree in point[" << j << "]: "
              << "(" << curves[2][j].xCoord()
              << "," << curves[2][j].yCoord() << ")"
              << " vs (" << curves[4][j].xCoord()
              << "," << curves[3][j].yCoord() << ")"
              << endl;
      }
   
   if (curves[2].size() != curves[3].size())
      cout << "ERROR: the two curves disagree in size. " << endl;
   
   ofstream outfile;
   Debug::initialize(argv[1]);

   vector< vector<Point> > b1_curve(1, curves[0]);
   vector< vector<Point> > b2_curve(1, curves[1]);
   vector< vector<Point> > or_curve(1, curves[2]);

   outfile.open("DEBUG/out0");
   Debug::PlotCurves(outfile, b1_curve);
//   Debug::PlotCurves(cout, b1_curve);   
   outfile.close();

   outfile.open("DEBUG/out1");
   Debug::PlotCurves(outfile, b2_curve);
//   Debug::PlotCurves(cout, b2_curve);
   outfile.close();

   outfile.open("DEBUG/out2");   
   Debug::PlotCurves(outfile, or_curve);
//   Debug::PlotCurves(cout, or_curve);
   outfile.close();

   cout << "Point::INFTY: " << Point::INFTY << endl;
   cout << "min_area of block A: " << BoundaryType(block_A).min_area()
        << " size: " << block_A.size() << endl;
   cout << "min_area of block B: " << BoundaryType(block_B).min_area()
        << " size: " << block_B.size() << endl;
   cout << "min_area[" << 0 << "]: " << BoundaryType(curves[0]).min_area()
        << " size: " << curves[0].size() << endl;
   cout << "min_area[" << 1 << "]: " << BoundaryType(curves[1]).min_area()
        << " size: " << curves[1].size() << endl;
   cout << "min_area[" << 2 << "]: " << BoundaryType(curves[2]).min_area()
        << " size: " << curves[2].size() << endl;
   cout << "min_point[" << 2 << "]: x:" << BoundaryType(curves[2]).min_point().xCoord()
        << " y:" << BoundaryType(curves[2]).min_point().yCoord() << endl;
}
// --------------------------------------------------------
void DebugHier(int argc, char *argv[])
{
   
   ifstream infile;
   ofstream outfile;

   infile.open(argv[1]);
   if (!infile.good())
   {
      cout << "ERROR: cannot open file: " << argv[1] << endl;
      exit(1);
   }

   outfile.open(argv[3]);
   if (!outfile.good())
   {
      cout << "ERROR: cannot open file: " << argv[3] << endl;
      exit(1);
   }

   int format = -1;
   if (!strcmp(argv[4], "--txt"))
      format = BlockInfoType::TXT;
   else if (!strcmp(argv[4], "--soft"))
      format = BlockInfoType::SOFT;

   cout << "---HIER MODE---" << endl;
   BlockInfoType blockinfo(format, infile);
   SoftHierEngineType engine(blockinfo);

   SoftSliceRecord *ssrPtr = engine();

   cout << "last records: " << endl;
   cout << "min_area: " << (ssrPtr->boundary).min_area() << endl;
   SoftPacking(*ssrPtr, blockinfo).output(outfile);
   
   outfile.close();
}
